package com.jiuqi.mde;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import com.jiuqi.mde.MDEDataInfo;
import com.jiuqi.mde.MDEDimValueSet;
import com.jiuqi.mde.MDEDimValueSetCollection;
import com.jiuqi.mde.MDEDimension;
import com.jiuqi.mde.MDEDimensionSet;
import com.jiuqi.mde.MDEException;
import com.jiuqi.mde.MDEField;
import com.jiuqi.mde.MDERecord;
import com.jiuqi.mde.MDERecordSet;
import com.jiuqi.mde.MDESqlLogger;
import com.jiuqi.mde.StatementCache;
import com.jiuqi.util.DBConnect;
import com.jiuqi.util.DuplicationException;
import com.jiuqi.util.IntValue;
import com.jiuqi.util.JqLib;
import com.jiuqi.util.SequencialList;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

public class MDETable extends MDERecordSet {
  public static boolean KEEP_FIELDS_ORDER = false;
  
  public boolean keepFieldsOrder;
  
  protected StatementCache stmtCache = new StatementCache();
  
  protected MDEDimValueSetCollection loadedDimValues = new MDEDimValueSetCollection();
  
  public MDETable(MDEDataInfo dataInfo) {
    super(dataInfo);
    this.keepFieldsOrder = KEEP_FIELDS_ORDER;
  }
  
  public MDERecord createRecord() {
    return new MDERecord(this.dataInfo);
  }
  
  public final MDERecord newRecord() throws MDEException {
    try {
      return newRecord(null);
    } catch (DuplicationException ex) {
      ex.printStackTrace();
      return null;
    } 
  }
  
  public final MDERecord newRecord(MDEDimValueSet dimValues) throws DuplicationException, MDEException {
    return newRecord(dimValues, true);
  }
  
  public final MDERecord newRecord(MDEDimValueSet dimValues, boolean checkNeedLoad) throws DuplicationException, MDEException {
    MDERecord result = null;
    if (getIsUseLoadDataSync()) {
      result = dvFind(dimValues);
      if (result == null)
        synchronized (this.mutexLoadData) {
          result = dvFind(dimValues);
          if (result == null)
            result = newRecord0(dimValues, checkNeedLoad); 
        }  
    } else {
      result = newRecord0(dimValues, checkNeedLoad);
    } 
    return result;
  }
  
  private MDERecord newRecord0(MDEDimValueSet dimValues, boolean checkNeedLoad) throws MDEException, DuplicationException {
    MDERecord result = createRecord();
    result.beforeNewRecord();
    if (dimValues != null)
      result.setDimValues(dimValues, null); 
    result.afterNewRecord();
    if (dimValues != null && checkNeedLoad && 
      !this.dataInfo.isTemporary() && needLoadData(result.getDimValues()))
      System.out.println(String.valueOf(this.dataInfo.getName()) + "应该首先载入数据，否则无法定位记录序号"); 
    try {
      add(result);
    } catch (DuplicationException ex) {
      if (this.hasOrder) {
        result.getDimValues().setValue("顺序号", null);
        add(result);
      } else {
        throw ex;
      } 
    } 
    result.calculateDefaultValues();
    return result;
  }
  
  public final List getRecords(DBConnect dbConnect, MDEDimValueSet dimValues) throws MDEException {
    loadData(dbConnect, dimValues);
    return dvFindMatches(dimValues);
  }
  
  public final MDERecord getRecord(DBConnect dbConnect, MDEDimValueSet dimValues, boolean createIfNotExist) throws MDEException {
    if (!dimValues.dimensions().equals(this.dataInfo.dimensions())) {
      MDEDimValueSet newDimValues = new MDEDimValueSet(this.dataInfo
          .dimensions());
      newDimValues.assignValues(dimValues);
      dimValues = newDimValues;
    } 
    MDERecord result = dvFind(dimValues);
    if (result == null) {
      List<MDERecord> records = getRecords(dbConnect, dimValues);
      if (records != null && records.size() > 1)
        throw new MDEException(String.valueOf(dimValues.toString()) + "维度编码不全，不能唯一定位" + 
            this.dataInfo.getName() + "记录"); 
      if (records != null && records.size() > 0)
        result = records.get(0); 
      if (result == null && (createIfNotExist || this.dataInfo.isTemporary()))
        try {
          result = newRecord(dimValues, false);
        } catch (DuplicationException ex) {
          ex.printStackTrace();
        }  
    } 
    return result;
  }
  
  public final boolean needLoadData(MDEDimValueSet dimValues) {
    return !this.loadedDimValues.isIncluded(dimValues);
  }
  
  public void dataLoaded(MDEDimValueSet dimValues) {
    this.loadedDimValues.include(dimValues);
  }
  
  public void loadData(DBConnect dbConnect, MDEDimValueSet dimValues) throws MDEException {
    if ((dimValues.dimensions().collectCompare(this.dataInfo.dimensions()) & 
      0x1) != 0) {
      MDEDimValueSet myDimValues = new MDEDimValueSet(this.dataInfo.dimensions());
      myDimValues.assignValues(dimValues);
      dimValues = myDimValues;
      dimValues.setValue("顺序号", null);
    } else if (dimValues.dimensions().hasOrder() && 
      dimValues.getValue("顺序号") != null) {
      dimValues = (MDEDimValueSet)dimValues.clone();
      dimValues.setValue("顺序号", null);
    } 
    if (!needLoadData(dimValues))
      return; 
    if (getIsUseLoadDataSync()) {
      synchronized (this.mutexLoadData) {
        if (!needLoadData(dimValues))
          return; 
        loadData0(dbConnect, dimValues);
      } 
    } else {
      loadData0(dbConnect, dimValues);
    } 
  }
  
  private void dealDuplication(MDERecord record) throws MDEException {
    MDERecord oldRecord = dvFind(record.getDimValues());
    MDEField field = this.dataInfo.getIndentyField();
    if (field != null && 
      JqLib.compareString((String)record.getFieldValue(field), 
        (String)oldRecord.getFieldValue(field)) != 0) {
      if (!this.dataInfo.dimensions().hasOrder()) {
        System.out.println(String.valueOf(this.dataInfo.getName()) + "数据库记录重码：" + 
            record.getDimValues().toString() + 
            "，记录被忽略");
      } else {
        System.out.println(String.valueOf(this.dataInfo.getName()) + "数据库记录重码：" + 
            record.getDimValues().toString() + 
            "，调整顺序号");
        record.getDimValues().setValue("顺序号", null);
        try {
          add(record);
        } catch (DuplicationException duplicationException) {}
        record.setDimValues(record.getDimValues(), null);
      } 
    } else {
      if ((oldRecord.recordFlag & 0x2) != 0)
        throw new MDEException(String.valueOf(this.dataInfo.getName()) + 
            "记录已经修改，不能再次从数据库载入"); 
      try {
        remove(oldRecord);
        add(record);
      } catch (DuplicationException duplicationException) {}
    } 
  }
  
  protected void add0(MDERecord record) throws MDEException {
    if (this.dimValuesIndex == null)
      buildDimValuesIndex(); 
    try {
      add(record);
    } catch (DuplicationException duplicationException) {
      dealDuplication(record);
    } 
  }
  
  private void loadData0(DBConnect dbConnect, MDEDimValueSet dimValues) throws MDEException {
    if (this.dimValuesIndex == null)
      buildDimValuesIndex(); 
    SequencialList result = loadFromTable(dbConnect, dimValues);
    for (int index = 0, count = result.size(); index < count; index++) {
      MDERecord record = (MDERecord)result.get(index);
      try {
        add(record);
      } catch (DuplicationException duplicationException) {
        dealDuplication(record);
      } 
    } 
    dataLoaded(dimValues);
  }
  
  public final void statModifies(List<MDERecord> modifyRecrods, MDEDimValueSetCollection modifyDimValues) {
    for (int i = 0, count = this.records.size(); i < count; i++) {
      MDERecord record = this.records.get(i);
      if (record.isInserted() || record.isDeleted() || record.isModified()) {
        if (modifyRecrods != null)
          modifyRecrods.add(record); 
        if (modifyDimValues != null)
          modifyDimValues.add(record.getDimValues()); 
      } 
    } 
  }
  
  public final void saveData(DBConnect dbConnect) throws MDEException {
    try {
      dbConnect.getConnection();
      try {
        this.stmtCache.clear();
        try {
          int i;
          int count;
          for (i = 0, count = this.records.size(); i < count; i++) {
            MDERecord record = this.records.get(i);
            if (record.isInserted())
              record.refreshDimValues(this.dimValuesIndex); 
            if (!record.isDeleting() && (
              record.isModified() || record.isInserted()) && 
              !record.hasData())
              record.delete(); 
          } 
          for (i = this.records.size() - 1; i >= 0; i--) {
            MDERecord record = this.records.get(i);
            if (record.isDeleting()) {
              deleteRecord(dbConnect, record);
              remove(i);
            } 
          } 
          for (i = 0, count = this.records.size(); i < count; i++) {
            MDERecord record = this.records.get(i);
            if (record.isInserted()) {
              insertRecord(dbConnect, record);
            } else if (!record.isDeleting()) {
              record.checkModification();
              if (record.isModified())
                updateRecord(dbConnect, record); 
            } 
          } 
        } finally {
          this.stmtCache.close();
        } 
      } finally {
        dbConnect.releaseConnection();
      } 
    } catch (SQLException ex) {
      throw new MDEException(ex);
    } 
  }
  
  public void resetRecords() {
    this.loadedDimValues.clear();
    super.resetRecords();
  }
  
  protected Object readDimension(ResultSet resultSet, MDEDimension dimension, IntValue index) throws SQLException {
    if ("顺序号" == dimension.name) {
      String order = null;
      if (index == null) {
        order = resultSet.getString(dimension.fieldName);
      } else {
        order = resultSet.getString(index.value++);
      } 
      try {
        return new Integer(JqLib.parseHexValue(order));
      } catch (NumberFormatException ex) {
        System.out.println(String.valueOf(this.dataInfo.getName()) + "顺序号无效");
        ex.printStackTrace();
        return new Integer(1);
      } 
    } 
    if (index == null) {
      if (dimension.isString)
        return resultSet.getString(dimension.fieldName); 
      return new Integer(resultSet.getInt(dimension.fieldName));
    } 
    if (dimension.isString)
      return resultSet.getString(index.value++); 
    return new Integer(resultSet.getInt(index.value++));
  }
  
  protected void writeDimension(PreparedStatement statement, MDEDimension dimension, IntValue index, Object value) throws SQLException {
    if ("顺序号" == dimension.name) {
      int order = (value != null) ? ((Integer)value).intValue() : 1;
      statement.setString(index.value++, JqLib.makeHexString(order));
    } else if (dimension.isString) {
      statement.setString(index.value++, (String)value);
    } else if (value instanceof Integer) {
      if (value == null) {
        statement.setNull(index.value++, 4);
      } else {
        statement.setInt(index.value++, ((Integer)value).intValue());
      } 
    } else if (value == null) {
      statement.setNull(index.value++, 2);
    } else {
      statement.setDouble(index.value++, ((Double)value).doubleValue());
    } 
  }
  
  protected Object readField(ResultSet resultSet, MDEField field, IntValue index) throws SQLException {
    if (index == null)
      return resultSet.getObject(field.name); 
    return resultSet.getObject(index.value++);
  }
  
  protected void writeField(PreparedStatement statement, MDEField field, IntValue index, Object value) throws SQLException {
    if (value == null) {
      statement.setNull(index.value++, field.dataType);
    } else {
      statement.setObject(index.value++, value);
    } 
  }
  
  protected final void readRecord(ResultSet resultSet, MDERecord record, boolean readDimension) throws SQLException {
    IntValue index = null;
    if (KEEP_FIELDS_ORDER)
      index = new IntValue(1); 
    if (readDimension) {
      MDEDimValueSet dimValues = new MDEDimValueSet(this.dataInfo.dimensions());
      for (int j = 0, k = this.dataInfo.dimensions().size(); j < k; j++) {
        MDEDimension dimension = this.dataInfo.dimensions().get(j);
        dimValues.setValue(j, readDimension(resultSet, dimension, index));
      } 
      record.setDimValues(dimValues, null);
    } 
    for (int i = 0, count = this.fields.size(); i < count; i++) {
      MDEField field = this.fields.get(i);
      record.setFieldValue(i, readField(resultSet, field, index));
    } 
  }
  
  protected final void writeFieldsData(PreparedStatement statement, MDERecord record, IntValue index) throws SQLException {
    for (int i = 0, count = this.fields.size(); i < count; i++) {
      MDEField field = this.fields.get(i);
      writeField(statement, field, index, record.getFieldValue(i));
    } 
  }
  
  protected final void writeDimensionsData(PreparedStatement statement, MDERecord record, IntValue index, boolean writeFieldDimension) throws SQLException {
    MDEDimValueSet dimValues = record.getDimValues();
    for (int i = 0, count = this.dataInfo.dimensions().size(); i < count; i++) {
      MDEDimension dimension = this.dataInfo.dimensions().get(i);
      if (writeFieldDimension || 
        this.fields.findByFieldName(dimension.fieldName) == null)
        writeDimension(statement, dimension, index, 
            dimValues.getValue(i)); 
    } 
  }
  
  protected final void writeRecord(PreparedStatement statement, MDERecord record, boolean writeDimension) throws SQLException {
    IntValue index = new IntValue(1);
    if (writeDimension) {
      MDEDimValueSet dimValues = record.getDimValues();
      for (int j = 0, k = this.dataInfo.dimensions().size(); j < k; j++) {
        MDEDimension dimension = this.dataInfo.dimensions().get(j);
        if (this.fields.findByFieldName(dimension.fieldName) == null)
          writeDimension(statement, dimension, index, 
              dimValues.getValue(j)); 
      } 
    } 
    for (int i = 0, count = this.fields.size(); i < count; i++) {
      MDEField field = this.fields.get(i);
      writeField(statement, field, index, record.getFieldValue(i));
    } 
  }
  
  private final void checkEmptyFields(MDEDimValueSet dimValues, ArrayList<MDEDimension> emptyFields) {
    for (int i = 0, count = dimValues.dimensions().size(); i < count; i++) {
      Object dimValue = dimValues.getValue(i);
      if (dimValue == null || "".equals(dimValue)) {
        MDEDimension dim = dimValues.dimensions().get(i);
        if (emptyFields.indexOf(dim) < 0)
          emptyFields.add(dim); 
      } 
    } 
  }
  
  private static final Integer DEFAULT_INT_VALUE = new Integer(999);
  
  private static final String DEFAULT_STR_VALUE = "_";
  
  private final void convertEmptyDimension(Connection conn, MDEDimension dim) throws SQLException {
    StringBuffer sql = new StringBuffer(80);
    sql.append("UPDATE ");
    sql.append(getTableName());
    sql.append(" SET ");
    sql.append(dim.fieldName);
    if (dim.isString) {
      sql.append("='_'");
    } else {
      sql.append("=99");
    } 
    sql.append(" WHERE ");
    sql.append(dim.fieldName);
    if (dim.isString) {
      sql.append("= '' OR ");
    } else {
      sql.append("=0 OR ");
    } 
    sql.append(dim.fieldName);
    sql.append(" IS NULL");
    Statement stmt = conn.createStatement();
    try {
      stmt.executeUpdate(sql.toString());
    } finally {
      stmt.close();
    } 
  }
  
  private final void changeEmptyDimensionValue(SequencialList records, MDEDimension dim) {
    for (int index = 0, count = records.size(); index < count; index++) {
      MDERecord rec = (MDERecord)records.get(index);
      MDEDimValueSet dimValues = rec.getDimValues();
      if (dimValues.getValue(dim) == null)
        if (dim.isString) {
          dimValues.setValue(dim, "_");
        } else {
          dimValues.setValue(dim, DEFAULT_INT_VALUE);
        }  
    } 
  }
  
  protected String convertSQL(String sql, MDEDimValueSet dimValues) {
    return sql;
  }
  
  protected void doLoadRecord(ResultSet rs, MDERecord record) throws SQLException, MDEException {
    record.beforeLoadRecord();
    readRecord(rs, record, true);
    record.afterLoadRecord();
  }
  
  private String generateSql(MDEDimValueSet dimValues) {
    return generateSql(dimValues, false);
  }
  
  private String generateSql(MDEDimValueSet dimValues, boolean isPrepared) {
    StringBuffer sql = null;
    if (!this.keepFieldsOrder) {
      sql = new StringBuffer(128);
      sql.append("SELECT * FROM ");
    } else {
      sql = new StringBuffer(this.fields.size() * 8 + 80);
      sql.append("SELECT ");
      int i, count;
      for (i = 0, count = this.dataInfo.dimensions().size(); i < count; i++) {
        sql.append((this.dataInfo.dimensions().get(i)).fieldName);
        sql.append(',');
      } 
      for (i = 0, count = this.fields.size(); i < count; i++) {
        MDEField field = this.fields.get(i);
        if (this.dataInfo.dimensions().findByFieldName(field.fieldName) == null) {
          sql.append(field.fieldName);
          sql.append(',');
        } 
      } 
      sql.deleteCharAt(sql.length() - 1);
      sql.append(" FROM ");
    } 
    sql.append(getTableName());
    sql.append(" WHERE ");
    sql.append(dimValues.getSQLFilter(isPrepared));
    return sql.toString();
  }
  
  protected void beforeReadRecord(MDEDimValueSet dimValues) {}
  
  protected void readRecord(ResultSet rs, SequencialList result, MDERecord record) throws SQLException, MDEException {
    doLoadRecord(rs, record);
    result.add(record);
  }
  
  protected void afterReadRecord(MDEDimValueSet dimValues) {}
  
  protected SequencialList loadFromTable(DBConnect dbConnect, MDEDimValueSet dimValues) throws MDEException {
    SequencialList result = new SequencialList();
    boolean isPrepared = false;
    String sql = generateSql(dimValues);
    try {
      Connection conn = dbConnect.getConnection();
      try {
        ArrayList<MDEDimension> emptyFields = new ArrayList();
        PreparedStatement stmt = null;
        try {
          beforeReadRecord(dimValues);
          String selectSQL = convertSQL(sql, dimValues);
          if (selectSQL.equals(sql)) {
            isPrepared = true;
            selectSQL = generateSql(dimValues, isPrepared);
          } 
          stmt = conn.prepareStatement(selectSQL);
          if (isPrepared)
            setParameter(stmt, dimValues); 
          ResultSet rs = stmt.executeQuery();
          try {
            while (rs.next()) {
              MDERecord record = createRecord();
              readRecord(rs, result, record);
              checkEmptyFields(record.getDimValues(), emptyFields);
            } 
            afterReadRecord(dimValues);
          } finally {
            rs.close();
          } 
          if (emptyFields.size() > 0)
            for (int i = 0, count = emptyFields.size(); i < count; 
              i++) {
              MDEDimension dim = emptyFields.get(i);
              convertEmptyDimension(conn, dim);
              changeEmptyDimensionValue(result, dim);
            }  
        } finally {
          stmt.close();
        } 
      } finally {
        dbConnect.releaseConnection();
      } 
    } catch (SQLException ex) {
      throw new MDEException(ex);
    } 
    return result;
  }
  
  private void setParameter(PreparedStatement ps, MDEDimValueSet dimValues) throws SQLException {
    if (dimValues == null)
      return; 
    MDEDimensionSet dimensions = dimValues.dimensions();
    Object[] values = dimValues.getDimValues();
    int index = 1;
    if (dimensions != null)
      for (int i = 0; i < dimensions.size(); i++) {
        MDEDimension dimension = dimensions.get(i);
        Object value = values[i];
        if (value != null)
          index = dimension.setParameter(ps, value, index); 
      }  
  }
  
  protected final boolean insertRecord(DBConnect dbConnect, MDERecord record) throws MDEException {
    try {
      int result = 0;
      record.beforeSaveRecord();
      try {
        result = doInsertRecord(dbConnect, record);
      } catch (SQLException e) {
        if (DBConnect.isUniqueViolationException(e)) {
          result = doUpdateRecord(dbConnect, record);
        } else {
          throw e;
        } 
      } 
      record.afterSaveRecord();
      if (MDESqlLogger.Sql_Logger != null)
        MDESqlLogger.Sql_Logger.LogInsertSql(makeInsertInfo(record, result)); 
      return (result > 0);
    } catch (SQLException ex) {
      throw new MDEException(ex);
    } 
  }
  
  private final int doInsertRecord(DBConnect dbConnect, MDERecord record) throws SQLException {
    String sqlString = null;
    PreparedStatement stmt = this.stmtCache.insertStatement;
    if (stmt == null) {
      StringBuffer sql = new StringBuffer(this.fields.size() * 10 + 100);
      StringBuffer param = new StringBuffer(20 + 2 * this.fields.size());
      sql.append("INSERT INTO ");
      sql.append(getTableName());
      sql.append(" (");
      int i, count;
      for (i = 0, count = this.dataInfo.dimensions().size(); i < count; i++) {
        MDEDimension dimension = this.dataInfo.dimensions().get(i);
        if (this.fields.findByFieldName(dimension.fieldName) == null)
          dimension.appendInsertSql(sql, param); 
      } 
      for (i = 0, count = this.fields.size(); i < count; i++) {
        sql.append((this.fields.get(i)).fieldName);
        sql.append(',');
        param.append("?,");
      } 
      sql.deleteCharAt(sql.length() - 1);
      param.deleteCharAt(param.length() - 1);
      sql.append(")");
      sql.append(" VALUES (");
      sql.append(param);
      sql.append(")");
      sqlString = convertSQL(sql.toString(), record.getDimValues());
    } 
    Connection conn = dbConnect.getConnection();
    try {
      if (stmt == null) {
        stmt = conn.prepareStatement(sqlString);
        this.stmtCache.insertStatement = stmt;
      } 
      IntValue index = new IntValue(1);
      writeDimensionsData(stmt, record, index, false);
      writeFieldsData(stmt, record, index);
      return stmt.executeUpdate();
    } finally {
      dbConnect.releaseConnection();
    } 
  }
  
  protected final boolean updateRecord(DBConnect dbConnect, MDERecord record) throws MDEException {
    try {
      int result = 0;
      record.beforeSaveRecord();
      result = doUpdateRecord(dbConnect, record);
      if (result == 0)
        result = doInsertRecord(dbConnect, record); 
      record.afterSaveRecord();
      if (MDESqlLogger.Sql_Logger != null)
        MDESqlLogger.Sql_Logger.LogInsertSql(makeUpdateInfo(record, result)); 
      return (result > 0);
    } catch (SQLException ex) {
      throw new MDEException(ex);
    } 
  }
  
  private final int doUpdateRecord(DBConnect dbConnect, MDERecord record) throws SQLException {
    MDEField identyField = this.dataInfo.getIndentyField();
    Object identyValue = (identyField == null) ? null : 
      record.getFieldValue(identyField);
    String sqlString = null;
    PreparedStatement stmt = (identyValue == null) ? 
      this.stmtCache.updateStatement : this.stmtCache.updateByIdentyStatement;
    if (stmt == null) {
      StringBuffer sql = new StringBuffer(this.fields.size() * 10 + 100);
      sql.append("UPDATE ");
      sql.append(getTableName());
      sql.append(" SET ");
      int i, count;
      for (i = 0, count = this.fields.size(); i < count; i++) {
        sql.append((this.fields.get(i)).fieldName);
        sql.append('=');
        sql.append("?,");
      } 
      sql.deleteCharAt(sql.length() - 1);
      sql.append(" WHERE ");
      if (identyValue != null) {
        sql.append(identyField.fieldName);
        sql.append("=?");
      } else {
        for (i = 0; i < this.dataInfo.dimensions().size(); i++) {
          MDEDimension dimension = this.dataInfo.dimensions().get(i);
          dimension.appendWhereClause(sql);
        } 
        sql.setLength(sql.length() - 4);
      } 
      sqlString = convertSQL(sql.toString(), record.getDimValues());
    } 
    Connection conn = dbConnect.getConnection();
    try {
      if (stmt == null) {
        stmt = conn.prepareStatement(sqlString);
        if (identyValue != null) {
          this.stmtCache.updateByIdentyStatement = stmt;
        } else {
          this.stmtCache.updateStatement = stmt;
        } 
      } 
      IntValue index = new IntValue(1);
      writeFieldsData(stmt, record, index);
      if (identyValue == null) {
        writeDimensionsData(stmt, record, index, true);
      } else {
        writeField(stmt, identyField, index, identyValue);
      } 
      return stmt.executeUpdate();
    } finally {
      dbConnect.releaseConnection();
    } 
  }
  
  protected final boolean deleteRecord(DBConnect dbConnect, MDERecord record) throws MDEException {
    try {
      record.beforeDeleteRecord();
      int result = doDeleteRecord(dbConnect, record);
      record.afterDeleteRecord();
      if (MDESqlLogger.Sql_Logger != null)
        MDESqlLogger.Sql_Logger.LogInsertSql(makeDeleteInfo(record.getDimValues(), result)); 
      return (result > 0);
    } catch (SQLException ex) {
      throw new MDEException(ex);
    } 
  }
  
  private final int doDeleteRecord(DBConnect dbConnect, MDERecord record) throws SQLException {
    MDEField identyField = this.dataInfo.getIndentyField();
    Object identyValue = (identyField == null) ? null : 
      record.getFieldValue(identyField);
    String sqlString = null;
    PreparedStatement stmt = (identyValue == null) ? 
      this.stmtCache.deleteStatement : this.stmtCache.deleteByIdentyStatement;
    if (stmt == null) {
      StringBuffer sql = new StringBuffer(this.fields.size() * 10 + 100);
      sql.append("DELETE FROM ");
      sql.append(getTableName());
      sql.append(" WHERE ");
      if (identyValue != null) {
        sql.append(identyField.fieldName);
        sql.append("=?");
      } else {
        for (int i = 0; i < this.dataInfo.dimensions().size(); i++) {
          MDEDimension dimension = this.dataInfo.dimensions().get(i);
          dimension.appendWhereClause(sql);
        } 
        sql.setLength(sql.length() - 4);
      } 
      sqlString = convertSQL(sql.toString(), record.getDimValues());
    } 
    Connection conn = dbConnect.getConnection();
    try {
      if (stmt == null) {
        stmt = conn.prepareStatement(sqlString);
        if (identyValue != null) {
          this.stmtCache.deleteByIdentyStatement = stmt;
        } else {
          this.stmtCache.deleteStatement = stmt;
        } 
      } 
      IntValue index = new IntValue(1);
      if (identyValue == null) {
        writeDimensionsData(stmt, record, index, true);
      } else {
        writeField(stmt, identyField, index, identyValue);
      } 
      return stmt.executeUpdate();
    } finally {
      dbConnect.releaseConnection();
    } 
  }
  
  public int deleteRecords(DBConnect dbConnect, MDEDimValueSet dimValues) throws SQLException {
    if (!dimValues.dimensions().equals(this.dataInfo.dimensions())) {
      MDEDimValueSet newDimValues = new MDEDimValueSet(this.dataInfo
          .dimensions());
      newDimValues.assignValues(dimValues);
      dimValues = newDimValues;
    } 
    String filter = dimValues.getSQLFilter();
    if (filter == null || filter.length() == 0)
      return 0; 
    Connection conn = dbConnect.getConnection();
    try {
      Statement stmt = conn.createStatement();
    } finally {
      dbConnect.releaseConnection();
    } 
  }
  
  protected String getTableName() {
    return this.dataInfo.getName();
  }
  
  private void writeFieldValues(MDERecord record, StringBuffer sb) {
    for (int i = 0, count = this.fields.size(); i < count; i++) {
      MDEField field = this.fields.get(i);
      sb.append('\t');
      sb.append(field.name);
      sb.append('=');
      Object value = record.getFieldValue(i);
      sb.append((value != null) ? value.toString() : "");
    } 
  }
  
  private String makeInsertInfo(MDERecord record, int retCode) {
    StringBuffer sb = new StringBuffer(2048);
    sb.append(MDESqlLogger.getLogTime());
    sb.append("\tINSERT\t");
    sb.append(getTableName());
    sb.append('\t');
    sb.append(retCode);
    sb.append('\t');
    sb.append(record.getDimValues().toString());
    writeFieldValues(record, sb);
    sb.append('\t');
    return sb.toString();
  }
  
  private String makeUpdateInfo(MDERecord record, int retCode) {
    StringBuffer sb = new StringBuffer(2048);
    sb.append(MDESqlLogger.getLogTime());
    sb.append("\tUPDATE\t");
    sb.append(getTableName());
    sb.append('\t');
    sb.append(retCode);
    sb.append('\t');
    sb.append(record.getDimValues().toString());
    writeFieldValues(record, sb);
    sb.append('\t');
    return sb.toString();
  }
  
  private String makeDeleteInfo(MDEDimValueSet dimValues, int retCode) {
    StringBuffer sb = new StringBuffer(2048);
    sb.append(MDESqlLogger.getLogTime());
    sb.append("\tDELETE\t");
    sb.append(getTableName());
    sb.append('\t');
    sb.append(retCode);
    sb.append('\t');
    sb.append(dimValues.toString());
    sb.append('\t');
    return sb.toString();
  }
}
